import {Context as HonoContext, Hono,} from "https://deno.land/x/hono@v4.3.11/mod.ts";
import {cors} from "https://deno.land/x/hono@v4.3.11/middleware.ts";
import {CashuMint, CashuWallet, getEncodedToken} from "@cashu/cashu-ts";
import {getAlias, getUserByNpub} from "./models.ts";
import {logger} from "./utils/index.ts";
import * as nip98 from "nostr-tools/nip98";
import {Client as LibSQL} from "@libsql/client";
import {npubToPubKeyString, pubKeyStringToNpub, TokenInfoWithMailSubscriptionDuration,} from "@arx/utils";

const NPUB_REGEX = /^npub1[023456789acdefghjklmnpqrstuvwxyz]{58}$/;
const CASHU_REGEX = /^cashu[A-Za-z0-9+-_]*={0,3}$/;

export class HttpServer {
  private app: Hono;

  constructor(private db: LibSQL, port: number) {
    this.app = new Hono();

    this.app
      .use("*", cors())
      .get("/", (c: HonoContext) => c.text("nostr.email server"))
      .get("/subscription/:npub", this.getSubscriptionForNpub)
      .get("/aliases/:npub", this.getAliasesForNpub)
      .get("/alias/:alias", this.getNpubForAlias)
      .post("/addAlias", this.addAlias)
      .post("/addTime/:npub", this.addTimeToNpub);

    Deno.serve({ port }, this.app.fetch);
    logger.info(`HTTP Server running on port ${port}`);
  }

  getSubscriptionForNpub = async (c: HonoContext) => {
    const npub = c.req.param("npub");
    if (!NPUB_REGEX.test(npub)) {
      return c.json({ error: "Invalid npub format" }, 400);
    }

    const user = await getUserByNpub(this.db, npub);

    if (!user) {
      return c.json({
        subscribed: false,
      });
    }
    return c.json({
      subscribed: true,
      subscribedUntil: user.subscriptionDuration == null
        ? Infinity
        : Math.floor(user.lastPayment.getTime() / 1000) +
          user.subscriptionDuration,
    });
  };

  getNpubForAlias = async (c: HonoContext) => {
    const aliasParam = c.req.param("alias");

    const alias = await getAlias(this.db, aliasParam);

    if (!alias) {
      return c.json({ error: "Not found" }, 404);
    }
    return c.json({ npub: alias.npub });
  };

  getAliasesForNpub = async (c: HonoContext) => {
    const npub = c.req.param("npub");
    if (!NPUB_REGEX.test(npub)) {
      return c.json({ error: "Invalid npub format" }, 400);
    }

    try {
      const unpacked = await this.getUnpackedAuthHeader(
        c.req.header("Authorization"),
        `/aliases/${npub}`,
      );
      const npubAsPubkey = npubToPubKeyString(npub);

      if (unpacked.pubkey !== npubAsPubkey) {
        return c.json({ error: "Unauthorized" }, 401);
      }

      const user = await getUserByNpub(this.db, npub);

      if (!user) {
        return c.json({ error: "Not found" }, 404);
      }

      return c.json(user.aliases);
    } catch (error) {
      if (error instanceof Error) {
        return c.json({ error: error.message }, 401);
      } else {
        return c.json({ error: `${error}` }, 401);
      }
    }
  };

  addAlias = async (c: HonoContext) => {
    const { alias } = await c.req.json<{
      alias: string;
    }>();

    try {
      const unpacked = await this.getUnpackedAuthHeader(
        c.req.header("Authorization"),
        "/addAlias",
      );
      const unpackedKeyToNpub = pubKeyStringToNpub(unpacked.pubkey);

      const user = await getUserByNpub(this.db, unpackedKeyToNpub);

      if (!user) {
        return c.json({ error: "Unauthorized" }, 401);
      }

      const stillHasSubscription = user.subscriptionDuration === null ||
        Math.floor(user.lastPayment.getTime() / 1000) +
              user.subscriptionDuration > Date.now() / 1000;

      if (!stillHasSubscription) {
        return c.json({ error: "User has no subscription" }, 400);
      }

      const aliasInDb = await getAlias(this.db, alias);

      if (aliasInDb) {
        return c.json({ error: "Alias already exists" }, 400);
      }

      await this.db.execute({
        sql: `
          INSERT INTO aliases (alias, npub)
          VALUES ($alias, $npub)
          ON CONFLICT (alias) DO NOTHING
        `,
        args: { alias, npub: unpackedKeyToNpub },
      });

      return c.json({ alias, npub: unpackedKeyToNpub });
    } catch (error) {
      if (error instanceof Error) {
        return c.json({ error: error.message }, 401);
      } else {
        return c.json({ error: `${error}` }, 401);
      }
    }
  };

  addTimeToNpub = async (c: HonoContext) => {
    const npub = c.req.param("npub");
    const { tokenString } = await c.req.json();

    if (!NPUB_REGEX.test(npub)) {
      return c.json({ error: "Invalid npub format" }, 400);
    }
    if (!CASHU_REGEX.test(tokenString)) {
      return c.json({ error: "Invalid Cashu token format" }, 400);
    }

    const user = await getUserByNpub(this.db, npub);

    if (
      user &&
      (user.subscriptionDuration === null ||
        user.subscriptionDuration === -1)
    ) {
      return c.json({ error: "User has unlimited subscription" }, 400);
    }

    const tokenInfo = new TokenInfoWithMailSubscriptionDuration(tokenString);
    const mint = new CashuMint(tokenInfo.mint);
    const wallet = new CashuWallet(mint);
    const newToken = await wallet.receive(tokenString);
    const encodedToken = getEncodedToken({
      mint: tokenInfo.mint,
      proofs: newToken,
    });
    logger.info(`New cashu token: ${encodedToken}`);
    if (user) {
      let timeRemaining = Math.max(
        0,
        Math.floor(
          (+new Date(
            user.lastPayment.getTime() +
              user.subscriptionDuration! * 1000,
          ) - +new Date()) / 1000,
        ),
      );
      timeRemaining += tokenInfo.duration;
      await this.db.execute({
        sql: `
          UPDATE users
          SET lastPayment = $lastPayment, subscriptionDuration = $subscriptionDuration
          WHERE npub = $npub
        `,
        args: {
          lastPayment: new Date(),
          subscriptionDuration: timeRemaining,
          npub,
        },
      });
      return c.json({
        newTimeRemaining: timeRemaining,
      });
    }
    await this.db.execute({
      sql: `
        INSERT INTO users (npub, registeredAt, lastPayment, subscriptionDuration)
        VALUES ($npub, $registeredAt, $lastPayment, $subscriptionDuration)
      `,
      args: {
        npub,
        registeredAt: new Date(),
        lastPayment: new Date(),
        subscriptionDuration: tokenInfo.duration,
      },
    });
    return c.json({
      newTimeRemaining: tokenInfo.duration,
    });
  };

  private getUnpackedAuthHeader = async (
    auth: string | undefined,
    url: string,
  ) => {
    if (!auth) {
      throw new Error("Unauthorized");
    }
    const authHeader = auth.split(" ")[1];
    const validate = await nip98.validateToken(
      authHeader,
      `${Deno.env.get("PUBLIC_API_BASE_URL")!}${url}`,
      "POST",
    );
    if (!validate) {
      throw new Error("Unauthorized");
    }
    return await nip98.unpackEventFromToken(authHeader);
  };
}
