import {Client as LibSQL} from "@libsql/client";

type User = {
  npub: string;
  registeredAt: Date;
  lastPayment: Date;
  subscriptionDuration: number | null;
};

type Alias = {
  alias: string;
  npub: string;
};

type UserWithAliases = User & {
  aliases: string[];
};

export type AliasRowResult = {
  rows: Array<{
    alias: string;
    npub: string;
  }>;
  columns: string[];
  rowsAffected: number;
};

export type UserRowResult = {
  rows: Array<{
    npub: string;
    registeredAt: Date;
    lastPayment: Date;
    subscriptionDuration: number | null;
    alias: string | null;
  }>;
  columns: string[];
  rowsAffected: number;
};

export const mapSingleAlias = (rows: AliasRowResult["rows"]): Alias | null => {
  if (rows.length === 0) {
    return null;
  }
  const firstRow = rows[0];
  return {
    alias: firstRow.alias,
    npub: firstRow.npub,
  };
};

export const mapSingleUserWithAliases = (
  rows: UserRowResult["rows"],
): UserWithAliases | null => {
  if (rows.length === 0) {
    return null;
  }
  const firstRow = rows[0];
  return {
    npub: firstRow.npub,
    registeredAt: firstRow.registeredAt,
    lastPayment: new Date(firstRow.lastPayment),
    subscriptionDuration: firstRow.subscriptionDuration,
    aliases: rows
      .filter((row) => row.alias !== null)
      .map((row) => row.alias!),
  };
};

export const getUserByNpub = async (
  db: LibSQL,
  npub: string,
): Promise<UserWithAliases | null> => {
  const result: UserRowResult = await db.execute({
    sql: `
        SELECT u.*, a.*
        FROM users u LEFT JOIN aliases a ON u.npub = a.npub
        WHERE u.npub = $npub;
    `,
    args: { npub },
  }) as any as UserRowResult;

  return mapSingleUserWithAliases(result.rows);
};

export const getUserByAlias = async (
  db: LibSQL,
  alias: string,
): Promise<UserWithAliases | null> => {
  const result: UserRowResult = await db.execute({
    sql: `
        SELECT u.*, a.*
        FROM users u LEFT JOIN aliases a ON u.npub = a.npub
        WHERE a.alias = $alias;
    `,
    args: { alias },
  }) as any as UserRowResult;

  return mapSingleUserWithAliases(result.rows);
};

export const getAlias = async (db: LibSQL, aliasParam: string) => {
  const result: AliasRowResult = await db.execute({
    sql: `
      SELECT *
      FROM aliases
      WHERE alias = $alias
      LIMIT 1;
    `,
    args: { alias: aliasParam },
  }) as any as AliasRowResult;

  return mapSingleAlias(result.rows);
};
