import {SMTPServer, SMTPServerAddress, SMTPServerDataStream, SMTPServerSession,} from "smtp-server";
import {deriveNsecForEmail, getNDK, getUserByAlias, logger,} from "./utils/index.ts";
import {NDKEvent, NDKKind, NDKPrivateKeySigner} from "@nostr-dev-kit/ndk";
import {Client as LibSQL} from "@libsql/client";
import {encryptEventForRecipient, parseEmail} from "@arx/utils";
import {concat} from "@std/bytes";
import path from "node:path";
import fs from "node:fs/promises";
import process from "node:process";

interface QueuedEmail {
  id: string;
  mailData: string;
  session: SMTPServerSession;
  attempts: number;
  createdAt: number;
}

export class NostrSmtpServer {
  private server: SMTPServer;
  private emailQueue: QueuedEmail[] = [];
  private isProcessing: boolean = false;
  private readonly MAX_RETRIES = 3;
  private readonly BACKUP_DIR = path.join(Deno.cwd(), "email-backups");

  constructor(private db: LibSQL, port: number) {
    this.server = new SMTPServer({
      authOptional: true,
      logger: false,
      onData: (
        stream: SMTPServerDataStream,
        session: SMTPServerSession,
        callback: () => void,
      ) => this.handleEmailData(stream, session, callback),
    });

    this.server.listen(port, "0.0.0.0");
    logger.info(`SMTP Server running on port ${port}`);

    fs.mkdir(this.BACKUP_DIR, { recursive: true }).catch((err) => {
      logger.error("Failed to create backup directory:", err);
    });

    this.setupGracefulShutdown();
    this.recoverFromBackups();
  }

  private async recoverFromBackups() {
    const files = await fs.readdir(this.BACKUP_DIR);
    for (const file of files) {
      try {
        const data = JSON.parse(
          await fs.readFile(path.join(this.BACKUP_DIR, file), "utf-8"),
        );
        this.emailQueue.push(data);
      } catch (error) {
        logger.error(`Failed to recover backup ${file}:`, error);
      }
    }
    await this.processQueue();
  }

  private setupGracefulShutdown(): void {
    const shutdown = async () => {
      logger.info("Graceful shutdown initiated");

      this.server.close();

      while (this.isProcessing) {
        await new Promise((resolve) => setTimeout(resolve, 100));
      }

      for (const email of this.emailQueue) {
        await this.backupEmail(email);
      }

      logger.info("Graceful shutdown completed");
      process.exit(0);
    };

    process.on("SIGTERM", shutdown);
    process.on("SIGINT", shutdown);
  }

  private async handleEmailData(
    stream: SMTPServerDataStream,
    session: SMTPServerSession,
    callback: () => void,
  ) {
    const chunks: Uint8Array[] = [];

    try {
      for await (const chunk of stream) {
        chunks.push(chunk);
      }

      if (!this.validateSender(session)) {
        callback();
        return;
      }

      const mailData = new TextDecoder().decode(
        concat(chunks),
      );

      const queuedEmail: QueuedEmail = {
        id: `${Date.now()}-${Math.random().toString(36).slice(2, 11)}`,
        mailData,
        session,
        attempts: 0,
        createdAt: Date.now(),
      };

      this.emailQueue.push(await this.backupEmail(queuedEmail));
      this.processQueue();
    } catch (e) {
      logger.error(`Error processing recipients: ${e}`, e);
    }

    callback();
  }

  private async backupEmail(email: QueuedEmail): Promise<QueuedEmail> {
    try {
      const backupPath = path.join(this.BACKUP_DIR, `${email.id}.json`);
      await fs.writeFile(backupPath, JSON.stringify(email));
    } catch (error) {
      logger.error("Failed to backup email:", error);
    }
    return JSON.parse(JSON.stringify(email)) as QueuedEmail; // returning the email because of a weird bug with smtp-server
  }

  private async processQueue(): Promise<void> {
    if (this.isProcessing) {
      return;
    }
    this.isProcessing = true;
    while (this.emailQueue.length > 0) {
      const email = this.emailQueue[0];
      if (email.attempts >= this.MAX_RETRIES) {
        this.emailQueue.shift();
        continue;
      }

      try {
        const parsedEmail: ReturnType<typeof parseEmail> = parseEmail(
          email.mailData,
        );
        await this.processRecipients(email.session, parsedEmail, this.db);

        // Remove from queue and delete backup if successful
        this.emailQueue.shift();
        await this.deleteBackup(email.id);
      } catch (error) {
        logger.error(`Failed to process email ${email.id}:`, error);
        email.attempts++;

        this.emailQueue.push(this.emailQueue.shift()!);
      }
    }
    this.isProcessing = false;
  }

  private async deleteBackup(id: string): Promise<void> {
    try {
      const backupPath = path.join(this.BACKUP_DIR, `${id}.json`);
      await fs.unlink(backupPath);
    } catch (error) {
      logger.error(
        `[TOXIC DATA!!!] Failed to delete email backup ${id}.json:`,
        error,
      );
    }
  }

  private validateSender(session: SMTPServerSession): boolean {
    if (!session.envelope.mailFrom) {
      logger.warn("Ignoring email without sender");
      return false;
    }
    return true;
  }

  private async processRecipients(
    session: SMTPServerSession,
    parsedEmail: ReturnType<typeof parseEmail>,
    db: LibSQL,
  ) {
    for (const recipientEmail of session.envelope.rcptTo) {
      const address = recipientEmail.address;
      const [alias, domain] = address.split("@");

      if (domain !== Deno.env.get("BASE_DOMAIN")) {
        logger.warn(
          `Not sending email to ${address} because it is not in the allowed domain`,
        );
        continue;
      }

      const user = await this.getUser(alias, db);
      if (!user || !this.isSubscriptionValid(user)) return;

      await this.sendNostrLetter(session, parsedEmail, user.npub);
    }
  }

  private async getUser(alias: string, db: LibSQL) {
    const user = await getUserByAlias(db, alias);

    if (!user) {
      logger.warn("No user found for", alias, "skipping");
      return null;
    }
    return user;
  }

  private isSubscriptionValid(
    user: NonNullable<Awaited<ReturnType<NostrSmtpServer["getUser"]>>>,
  ): boolean {
    // If there's no duration set, it's an unlimited subscription
    if (user.subscriptionDuration === null) {
      return true;
    }

    const subscriptionDurationMs = user.subscriptionDuration * 1000;
    const lastPaymentTimestamp = user.lastPayment.getTime();
    const currentTimestamp = Date.now();

    const subscriptionEndTime = lastPaymentTimestamp + subscriptionDurationMs;
    const timeRemaining = subscriptionEndTime - currentTimestamp;

    if (timeRemaining <= 0) {
      logger.warn(`Subscription has expired for ${user.npub}`);
      return false;
    }
    return true;
  }

  private async sendNostrLetter(
    session: SMTPServerSession,
    parsedEmail: ReturnType<typeof parseEmail>,
    recipient: string,
  ) {
    const randomKeySinger = new NDKPrivateKeySigner(
      deriveNsecForEmail(
        Deno.env.get("MASTER_NSEC")!,
        (session.envelope.mailFrom as SMTPServerAddress).address,
      ),
    );

    const ndk = getNDK();
    ndk.signer = randomKeySinger;
    await ndk.connect();

    const ndkUser = ndk.getUser({ npub: recipient });
    const randomKeyUser = await randomKeySinger.user();
    const event = new NDKEvent();
    event.kind = NDKKind.Article;
    event.content = parsedEmail.body;
    event.created_at = Math.floor(Date.now() / 1000);
    event.pubkey = randomKeyUser.pubkey;
    event.tags.push(
      ["p", ndkUser.pubkey],
      ["subject", parsedEmail.subject],
      ["email:localIP", session.localAddress],
      ["email:remoteIP", session.remoteAddress],
      ["email:isEmail", "true"],
      ["email:session", session.id],
      ["email:from", (session.envelope.mailFrom as SMTPServerAddress).address],
    );

    for (const to of session.envelope.rcptTo) {
      event.tags.push(["email:to", to.address]);
    }

    for (const header of Object.keys(parsedEmail.headers)) {
      event.tags.push([`email:header:${header}`, parsedEmail.headers[header]]);
    }

    await event.sign(randomKeySinger);
    const encryptedEvent = await encryptEventForRecipient(ndk, event, ndkUser);
    await encryptedEvent.publish();
  }
}
